# -*- coding: utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

__author__ = 'Adam Johnson'
__email__ = 'me@adamj.eu'
__version__ = '2.5.0'


default_app_config = 'django_mysql.apps.MySQLConfig'
